/*
 *  "sertest.c"  -  Test I/O program to see if a user 68000 program can keep up
 *		    with characters coming in on a serial channel at line rate
 *
 *    08/21/1984	John Elliott IV
 * ============================================================================
 *
 *	To compile:	"cc -O sertest.c -o sertest"
 *
 *	To run:		"sertest <ttyxx> [baud] [read-size]"
 *			Example:  "sertest tty02 4800 3" will run the test on
 *			tty02 at 4800 baud with a read buffer of 3 bytes.
 *			<baud> defaults to 4800 baud, and the default read-size
 *			is a buffer of 256 bytes.
 *
 *		 Note:  On the Host end (the VAX) make sure you are logged in at
 *			the specified baud rate and have a file in the current
 *			working directory called "testfile", which contains
 *			as many or as few of the following lines as you wish:

0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~

 *			It is an interesting test of the operating system's
 *			buffering capacity to run sertest with a fairly large
 *			"testfile" at a fast (say, 4800) baud rate.
 */

#include  <stdio.h>
#include  <sgtty.h>
#include  <signal.h>

#define	error(x)	fprintf(stderr, x)
#define	error1(x, a1)	fprintf(stderr, x, a1)
#ifdef DEBUG
#  define debug(x)		fprintf(stderr, x)
#  define debug1(x, a1)		fprintf(stderr, x, a1)
#  define debug2(x, a1,a2)	fprintf(stderr, x, a1,a2)
#else DEBUG
#  define debug(x)
#  define debug1(x, a1)
#  define debug2(x, a1,a2)
#endif DEBUG


/*
 *  Constants, globals, and other fun stuff
 */
#define	MaxREADSZ 1024

char	startcmd[] = "\025gets >& /dev/null; stty nl; cat testfile; echo '*-EOF-*'\n";
char	stopcmd[] = "\003\003\003 stty -nl\n";
char	eof[] = "*-EOF-*";
char	go[] = "\n";

int	gfd, Readsz;
int	nreads = 0, nbytesrd = 0;


main(argc, argv)
  int	argc;
  char	**argv;
{
	register char	*tty, *ptr, dev[20], buff[512];
	register int	fd, lines;
	struct	 sgttyb	opts;
		 int	speed;
		 char	*see();
		 int	serintr();

	--argc, ++argv;
	if (argc < 1) {
	    error("Missing input device.\n");
	    usage(1);
	}
	if (argc > 3) {
		error("Too many arguments.\n");
		usage(1);
	}

	/*
	 *  Build the filename
	 */
	tty = *argv;
	--argc, ++argv;
	sprintf(dev, "/dev/%s", tty);

	/*
	 *  Get the line rate, if specified.  Default to 4800 baud.
	 */
	if (argc > 0) {
	    speed = atoi(*argv);
	    switch (speed) {
	        case 300:
		    speed = B300;
		    break;
	        case 600:
		    speed = B600;
		    break;
	        case 1200:
		    speed = B1200;
		    break;
	        case 2400:
		    speed = B2400;
		    break;
	        case 4800:
		    speed = B4800;
		    break;
	        case 9600:
		    speed = B9600;
		    break;
	        default:
	    	    error1("Bad line speed \"%s\" specified.\n", *argv);
	    	    error("Use 300, 600, 1200, 2400, 4800, or 9600 baud.\n");
		    exit(1);
	    }
	    --argc, ++argv;
	} else
		speed = B4800;

	if (argc > 0) {
		Readsz = atoi(*argv);
		if (Readsz < 0  ||  Readsz > MaxREADSZ) {
			error("Bad read buffer size.  ");
			error1("Use a value between 1 and %d bytes.\n",
								MaxREADSZ);
			exit(1);
		}
		--argc, ++argv;
	} else
		Readsz = 256;

	/*
	 *  Try to open the device and set the speed and modes
	 */
	gfd = fd = open(dev, 2);
	if (fd < 0) {
		error("sertest:  Can't open ");
		perror(dev);
		exit(1);
	}
	gtty(fd, &opts);
	opts.sg_ispeed = opts.sg_ospeed = speed;	/* set the line speed */
	opts.sg_flags |= CBREAK;
	opts.sg_flags &= ~ECHO;			/* This is *VERY* important */
	opts.sg_flags &= ~CRMOD;
	stty(fd, &opts);

	signal(SIGINT, serintr);

	/*
	 *  Output the starting command to the host
	 */
	if (write(fd, startcmd, sizeof(startcmd)) < sizeof(startcmd)) {
		perror("write(startcmd) error");
		write(fd, stopcmd, sizeof(stopcmd));
		close(fd);
		exit(1);
	}
	/*
	 *  Verify and throw away the echoed command
	 */
	tgets(buff, sizeof(buff), fd);
	printf("%s", buff);

	/*
	 *  Read until the stop sequence received, or until error
	 */
	lines = 0;
	printf("[start]\n");
	write(fd, go, sizeof(go));	/* send a ^J to exit gets cmd */

	/* eat any garbage echoed back after the START (^J) request */
	tgets(buff, sizeof(buff), fd);	/* Eat the cr/lf echoed back */
	debug1("%s", buff);
	while (*buff != '0') {
		tgets(buff, sizeof(buff), fd);
		debug1("%s", buff);
#ifndef DEBUG
		putc('+', stderr);
#endif DEBUG
	}

	for ( ; ; ) 
	{
		int	size;

		size = tgets(buff, sizeof(buff), fd);
		debug1("%s", buff);

		if (size < 80) {
		    if (strncmp(buff, eof, sizeof(eof) - 1) == 0) {
			error("\n*** End of file reached.\n");
			break;
		    }	
		    error("\n*** Receive read error!  Aborting...\n");
		    printf("[got %d chars, wanted 80]\n", size);
		    buff[strlen(buff) - 1] = '\0';
		    printf("[buff was \"%s\"]\n", buff);
		    break;
		}

		{
		    register char  want = '0';

		    for (ptr = buff,want = '0'; ptr < buff+79; ++ptr,++want) {
			if (*ptr != want) {
			    error("\n*** Receive char error!  Aborting...\n");
			    printf("[offset %d, char was %s, wanted '%c']\n",
				  ptr - buff, see((unsigned char)*ptr), want);
			    break;
			}
		    }
		}
		if (ptr < buff + 79)
			break;
		if (*ptr != '\n') {
			error("\n*** Receive char error!  Aborting...\n");
			printf("[offset %d, char was %s, wanted '^J']\n",
					ptr - buff, see((unsigned char)*ptr));
			break;
		}
		++lines;
#ifndef DEBUG
		putc('.', stderr);
#endif DEBUG
	}

      /*
       *  Clean up and show statistics
       */
      printf("[%d lines read, %d bytes in %d reads; averaged %f bytes/read]\n",
			lines, nbytesrd, nreads, (float)nbytesrd/(float)nreads);
      write(fd, stopcmd, sizeof(stopcmd));
      close(fd);
      exit(0);
}


tgets(buff, size, fd)
  register char	*buff;
  register int	fd;
	   int	size;
{
	static	 char	rxbuff[MaxREADSZ];
	static	 char	*current_ptr;
	static	 int	nbytes = 0;

	register char	*ptr;
	register int	bytes, nread;
	

	bytes = size;
	nread = nbytes;
	ptr = current_ptr;

	while (bytes) {
		if (nread <= 0  ||  ptr - rxbuff >= nread) {
			nread = nbytes = read(fd, rxbuff, Readsz);

			nbytesrd += nread;
			++nreads;

			if (nread <= 0) {
				debug1("tgets:  read returned %d\n", nread);
				error("tgets:  read() error; ");
				perror("");
				break;
			} else
				ptr = rxbuff;
		}
		--bytes;
		if ((*buff++ = *ptr++) == '\n')
			break;
	}

	*buff = '\0';
	current_ptr = ptr;
	return (size - bytes);
}

char *
see(c)
  register unsigned char c;
{
	static	 char	buff[10];

	if (c >= ' ' && c <= '~')
		sprintf(buff, "'%c'", c);
	else if (c < ' ')
		sprintf(buff, "\"^%c\"", (c | '@'));
	else
		sprintf(buff, "0x%x", c);

	return(buff);
}

serintr()
{
	signal(SIGINT, SIG_IGN);
	error("\n[break:  Aborting...]\n");
	write(gfd, stopcmd, sizeof(stopcmd));
	close(gfd);
	exit(1);
}

usage(code)
{
	error("Usage:  sertest <ttyxx> [speed] [read-size (bytes)]\n");
	exit(code);
}
