/*********************************

z80disk.h -- interface to z80 disks

Copyright (C) Tandy Corporation, 1985.

This file must be kept consistent with z80/ctl/68kdefs.h and z80/ctl/bss.s

*********************************

The z80 floppy driver determines the format of a disk at mount time
and sets the parameter entries in the zd_iopb structure appropriately.
Floppies can be single or double sided, single or double density,
with an optional track0 side 0 IBM 128 byte 26 sector single density
boot track.  The formats supported include, for one or two sides:
	single density:
		26 128 byte sectors (IBM SD)
		16 256 byte sectors
		8 512 byte sectors
		4 1024 byte sectors
	double density:
		52 128 byte sectors (performance not guaranteed)
		26 256 byte sectors (TRSDOS II)
		16 512 byte sectors (XENIX, TRSDOS IV)
		8 1024 byte sectors

If the iopb.zd_fmt & MIXED_FMT is set, any ref to cyl 0 track 0 is implicitly 
in 128 byte sectors.  

XENIX has one driver for all the disks, but the driver knows it has
separate controllers for the floppies and the winnies.

The z80 can do multiple sector continuous I/O at the full raw speed
of the disk (40,000 bytes/DD, 20,000 bytes/SD).  This is important for
swapping and other raw I/O.  XENIX calculates the starting cyl/track/sector
for each transfer and gives a sector count; the z80 carries the count
across multiple tracks and cylinders as per the size params in zd_iopb.

Scenario for disk I/O:
	68k writes cmd, dma address, count etc. to iopb and sets GO
	z80 sees GO [copies write data to z80], starts disk I/O 
	disk done interrupts z80
	[z80 copies read data to 68k]
	z80 updates iopb, clears GO, sets INTPEND, [interrupts 68k]
	68k polls INTPEND for each disk iopb; for each one set it
		clears it and calls the intrp routine for that drive.

The 68k issues a DINIT command at every first-open for a drive.  The z80
then determines the disk format, sets the iopb params as per the disk 
format and interrupts the 68k.  The z80 will give an error if the disk 
is changed without a DINIT before the next read/write cmd.

The commands DREAD and DWRITE read and write data. DWRITEV writes the data,
then verifies it and returns status as if a read command had been issued.

****/

struct zd_iopb {	/* one structure per real controller */
/* these are set by the 68k for each transfer */
	caddr_t	zd_addr;	/* 68k data address, must be first entry */
	unsigned short
		zd_count;	/* SECTOR count (not BLOCKS) */
	union {
	    struct {
	   	unsigned short
			cyl,	/* starting cyl */
			hd,	/* starting head (side) */
			sec;	/* starting sector (NOT MULTIPLE OF BSIZE) */
	    } chs;
	    daddr_t	dblk;	/* Starting disk-relative-block for IOmega */
	} zd_loc;
/* these are set by the z80 on a DINIT and used by the 68k to calculate the 
   starting cyl, track and sector for a transfer.  */
	unsigned short
		zd_ncyl,	/* cyls / disk, set by z80 on INIT */
		zd_secsiz,	/* bytes / sector */
		zd_trksiz;	/* sectors / trk */
	unsigned char
		zd_nhead,	/* heads / cyl ((SIDES for floppies)) */
		zd_fmt;		/* density, mixed sd-dd format, write-protect */
/* these are set by 68k and z80 for each cmd. */
	unsigned char
		zd_csr,		/* command/status register */
		zd_xcsr,	/* one's complement of zd_csr for safety */
		zd_dev,		/* drive number */
		zd_err,		/* error code if zd_csr&DERROR */
		zd_seek,	/* seek rate (floppy, 0..3) */
		zd_mode;	/* write/verify and floppy drive mode */
	unsigned short
		zd_diag;	/* supplemental error info */
};

/* How to get at everything you need in zd_loc: */
#define	zd_cyl	zd_loc.chs.cyl
#define	zd_hd	zd_loc.chs.hd
#define	zd_sec	zd_loc.chs.sec
#define	zd_blk	zd_loc.dblk

/* bits in zd_fmt -- Set by the z80 at INIT */
#define	MIXED_FMT	0x01	/* cyl 0 hd 0 is IBM-SD, rest is per params */
#define DBL_DENS	0x02	/* disk is double-density */
#define WRPROT		0x10	/* write protected, also defined in zd_err */

/* bits in zd_csr --
	To start an operation the 68k first writes the one's complement
	of the csr bits to zd_xcsr and then writes the bits to zd_csr.
	The 68k can't touch the csr after it sets GO until the z80 clears GO.
	The z80 can't touch the csr after it clears GO until the 68k sets GO.
 */
#define	GO	0x01	/* set by 68k to begin xfer, cleared by z80 when done */
#define	INTENAB	0x02	/* set by 68k to enable interrupt on -GO */

#define	INIT	0x00	/* set by 68k to get disk geometry and mount/unmount */
#define	READ	0x04	/* set by 68k for disk read */
#define	WRITE	0x08	/* set by 68k for disk write */
#define FORMAT	0x10	/* set by 68k for disk device format (if available) */
#define	UMOUNT	0x80	/* When set and cmd == INIT, this is a disk UNMOUNT;
			   when OFF and cmd == INIT, this is a MOUNT */

#define	INTPEND	0x20	/* set by z80 when interrupting 68k, 68k can clear it
				to acknowledge interrupt */
#define	ERROR	0x40	/* set by z80 if and error occured, this and the error
				code in zd_err are valid from when the z80 
				clears GO to when 68k sets GO */

/* bits in zd_mode -- set by 68k */
#define	WRVERIFY  0x01	/* verify after write */
#define	MOTORWAIT 0x02	/* wait 2 sec instead of instant not-ready error */
#define	DOOROPEN  0x04	/* don't ignore door-opened status from drive */

/* bits in zd_err -- set only by z80 */
#define	SOFTERR	 0x01	/* required retries BUT THE DATA IS VALID */
#define	HARDERR	 0x02	/* unrecoverable r/w or drive error */
#define	NOTREADY 0x04	/* drive not ready  */
#define	INVALID	 0x08	/* disk change */
/*#define WRPROT 0x10	/* drive is write-protected (defined in zd_fmt also) */
#define CMDERR	 0x20	/* 68k command error */
#define BADFORM	 0x40	/* Disk has no .pvh or is unreadable */

/* the actual iopb's */
struct zd_iopb z80disk[3];
