/*****************************************

z80ser.h -- interface to z80 serial lines

Copyright (C) Microsoft Corporation, 1982.

This file must be kept consistent with z80/ctl/68kdefs.h and z80/ctl/bss.s

******************************************

The z80 serial interface has the functionality of a DEC DH-11.

Scenario for serial output:
Each line has a dma address and count, a baud rate, some mode bits and a csr.
	The 68k sets up everything and then sets GO and INTENAB in the csr
		and returns awaiting the interrupt.
	The z80 starts output and takes one char from the 68k at each tx intr.
	When the output queue is empty, the z80 clears GO and sets
		TXINT and, if TXINTENAB is set, interrupts the 68k.
	The 68k clears TXINT and calls the Xenix transmit 
		intr routine for each line that has TXINT set 
		(and also calls the Xenix receive intr routine which 
		processes any chars that are in the receive fifo)

The z80 checks the mode bits for a transmit stop before sending each char.
When the z80 detects TXSTOP, it sets zd_count to the number of chars sent 
so far.  The baud rate is updated in the polling loop; at worst at every GO.

Scenario for serial input:
The z80 receives characters and places them on the input fifo; the z80
has an input pointer and the 68k has an output pointer.  The polling
loop checks the fifo alarm and number of characters and interrupts the 68k
if required.  The fifo alarm level is updated each second.  The interrupt is 
immediate when the lines are idle and every 32 characters when the system 
is receiving lots of input.
The z80 checks RXDISAB as each char is received and stores the chars
in the fifo if RXDISAB is not set.

XENIX is fast enough to read all received chars from the fifo
and deliberately throw them away.  The z80 enters OVERRUN in the fifo
if either the hardware or the fifo overflows.

Modem Control:
The z80 polls for changes in zs_wctl and updates the output lines for
the device if they change.  The serial device status interrupts update
zs_rctl, which may be polled by the 68k.  If MINTENAB is set, the z80
interrupts the 68k by passing a character in the fifo with the MCTLNTFY
status bit set.  Also, TXBREAK is now implemented.
****/

/* communication block for z80 serial lines, one per line */

struct zs_iopb {
	caddr_t	zs_addr;	/* address of character list, set by 68k */
	unsigned char
		zs_csr,		/* control and status, set by z80 and 68k */
		zs_count,	/* size of character list, set by 68k */
		zs_baud,	/* baud rate, set by 68k */
		zs_mode,	/* mode, set by 68k */
		zs_wctl,	/* modem control outputs set by 68k */
		zs_rctl;	/* modem control inputs read by 68k */
};

/* communication block for received character fifo, one only */
    /*
     *  WARNING:  The Z80 code now *depends* on the fact that FIFOSIZE
     *	      is now 256, as now it can keep the array index in a 
     *	      in a register and not worry about modulo arithmatic
     */
#define FIFOSIZE 256
struct zs_rcv {
	unsigned char
		zr_inptr,  /* input ptr, set by z80 [index into zs_fifo] */
		zr_outptr, /* output ptr, set by 68k */
		zr_csr;	   /* control and status, set by z80 and 68k */
}; /* this is really one structure but it's messy to put the fifo inside. */
struct zs_fifo {
	unsigned char
		zf_s,	/* line number and status for each char */
		zf_ch;	/* each char */
};			/* struct zs_fifo z80fifo[FIFOSIZE]; */

/* bits in zs_csr 
   68k can't modify zs_csr after setting GO until z80 clears GO
   z80 can't modify after clearing GO until 68k sets GO
 */
#define	TXGO	0x01	/* 68k sets to start output, z80 clears when done */
#define	TXIENAB	0x04	/* 68k sets to enable interrupt on -GO */
#define	TXINT	0x10	/* z80 sets when done, 68k may reset for its own ack */

/* bits in zs_baud -- set only by 68k */
	/* The z80 only reads this byte if we do a RqSTwrite z80attn() call */
#define	BAUDMSK	0x0F	/* These bits are set to  B0 - B9600  */
#define	WORDMSK	0x30	/* These bits are set as below:    */
#define	WORD5	0x00	/* This yields a 5 bit word length */
#define	WORD6	0x10	/* This yields a 6 bit word length */
#define	WORD7	0x20	/* This yields a 7 bit word length */
#define	WORD8	0x30	/* This yields a 8 bit word length */

/* bits in zs_mode -- set only by 68k */
	/* The z80 only reads these bits if we do a RqSTwrite z80attn() call */
#define	RXDISAB	0x01	/* inhibit receiver on this line */
#define	TXSTOP	0x02	/* stop and flush output, set zs_count = n sent */

	/* The z80 checks these bits in the polling loop */
#define	TXBREAK	0x10	/* transmit break level */
#define PARENAB 0x20	/* enable transmission and checking of parity */
#define zPARODD	0x40	/* odd parity if enabled */
#define STOP2	0x80	/* transmit two stop bits */

/* bits in zs_rctl -- control information register [set only by z80] */
#define DCDON	0x01	/* indicates DCD (Data Carrier Detect) is asserted */
#define CTSON	0x02	/* indicates CTS (Clear To Send) is asserted */
#define DSRON	0x04	/* indicates DSR is asserted */

/* bits in zs_wctl -- control command register [set only by 68k] */
#define RTSON	0x01	/* set to assert RTS (Request To Send) */
#define DTRON	0x02	/* set to assert DTR (Data Terminal Ready) */
#define MINTENAB 0x04	/* set to enable status change interrupts */

/* bits in zr_csr - receiver control and status byte */
	/* WARNING: a race condition can occur if XENIX ever resets RXIENAB */
#define	RXIENAB	0x04	/* 68k sets to enable receiver interrupts */
#define	RXINT	0x10	/* z80 sets on interrupt, 68k clears for ack */

/* bits in zf_s - status byte for each char read from fifo */
#define	LINENO	 0x0f	/* mask for line number */
#define	PARERR	 0x10	/* set on parity error */
#define	zOVRRUN	 0x20	/* set if data was lost due to fifo overflow */
#define	FRAMERR	 0x40	/* set on framing error (received break) */
#define MCTLNTFY 0x80	/* set if this is a dummy char to notify of ctl chnge */

/* The actual io structures */
#define	NZ80SER	13			/* number serial channels */
struct zs_iopb z80ser[NZ80SER];

#define	NZ80LP	1		/* number of line-printer channels */
/* struct zs_iopb z80lp[NZ80LP]; */

struct zs_rcv  z80rcv;			/* 2 -> 4 bytes */
struct zs_fifo z80fifo[FIFOSIZE];	/* 512 bytes */
