/*
 *	@(#) sys/locking.c 1.1 83/10/23 Copyright Microsoft Corporation, 1983
 *
 *  $Header: locking.c,v 6.0 87/10/12 17:09:21 iv Exp $
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */


/*
 * file lock routines
 * John Bass, PO Box 1223, San Luis Obispo, CA 93401
 * original design spring 1976, CalPoly, San Luis Obispo
 * Deadlock idea from Ed Grudzien at Basys April 1980
 * extensions fall 1980, Onyx Systems Inc., San Jose
 *        Locking is called thru sysent. The arguments are assembled and
 *   verified. If in error, the request is aborted.
 *   If the request is unlock, then deletion rules are used to
 *   modify or remove the lock on the requested regions.
 *   If the request is to lock, then locked is called with the type
 *   of the lock request. If locked returns an error the lock request is
 *   aborted. Otherwise insertion rules are used to lock the requested
 *   region, and requests are compacted if necessary.
 *
 *   Flag values:
 *              LK_UNLCK  0       unlock request
 *              LK_LOCK   1       lock request
 *              LK_NBLCK  2       non-blocking lock request
 *              LK_RLCK   3       read permitted lock request
 *              LK_NBRLCK 4       non-blocking read permitted lock request
 *
 *      Modifications
 *      M000    28 Dec 81       RWS
 *      	- added read-permitted locks
 *	M001	01 Nov 82	JJD
 *		- eliminated unused local variable "foo" in locking
 *		- cast arg to wakeup to be caddr_t
 *		- fixed locked so that it properly returns a value (0) when
 *		  read is being attempted on a read-only locked region
 *		- locked returns 0 instead of NULL
 *	M002	04 Nov 82	EA
 *		- fixed locking so that the test to prevent trying to lock
 *		  a directory uses i_mode and not i_flag.
 *	M003	12 Feb 83	JJD
 *		- added some casts for lint, made necessary by recent change
 *		  of ll_flags field of struct locklist to unsigned short.
 *		  (Cast ll_flags to int for call to lockadd.)  (Uncommented.)
 *
 * $Log:	locking.c,v $
 * Revision 6.0  87/10/12  17:09:21  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.0  85/08/23  10:34:46  iv
 * 	- Checkin to lock down XENIX 3.1 release
 * 
 * Revision 4.0  85/03/13  14:47:58  iv
 * 	- Checkin to lock down XENIX 3.0(28) release
 * 
 * Revision 3.0  85/01/17  15:43:23  iv
 * 	- Checkin to save XENIX v3.0(23)
 * 
 * Revision 2.0  85/01/10  13:22:29  iv
 * Checkin to save XENIX 3.0(22)
 * 
 * Revision 0.1  84/05/17  15:13:13  iv
 * 	- No code changes.  Added RCS stuff.
 * 
 */

#include "../h/param.h"
#include "../h/inode.h"
#include "../h/file.h"
#include "../h/proc.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/systm.h"
#include "../h/locking.h"
#include "../h/var.h"

#define MAXSIZE (long)(1L<<30)     /* number larger than any request */

/*
 * locking -- handles syscall requests
 */

locking() {
     struct file *fp, *getf();
     struct inode *ip;
     /*
      * define order and type of syscall args
      */
     register struct a {
	  int  fdes;
	  int  flag;
	  off_t     size;
     } *uap;
     register struct locklist *cl, *nl;
     off_t     LB, UB;  /* note: UB is the first byte NOT covered by the lock! */
     int       nblock,rolock;
     int       oldf;
     off_t     oldend;

     /*
      * check for valid open file
      */
     uap = (struct a *)u.u_ap;
     fp = getf(uap->fdes);
     if(fp == NULL) return;
     ip = fp->f_inode;
     if ((ip->i_mode&IFMT) == IFDIR) {			/* M002 */
	  u.u_error = EACCES;
	  return;
     }

     /*
      * validate ranges
      * kludge for zero length
      */
     LB = fp->f_offset;
     if( uap->size != 0 ) {
	  UB = LB + uap->size;
	  if(UB <= 0) UB = MAXSIZE;
     }
     else UB = MAXSIZE;

     nblock = (uap->flag==LK_NBLCK || uap->flag==LK_NBRLCK) ? 2 : 0; /* M000 non blocking? */
     rolock = (uap->flag== LK_RLCK || uap->flag==LK_NBRLCK) ? 1 : 0; /* M000 read permitted? */
     /*
      * test for unlock request
      */
     if(uap->flag == 0) {
	  /*
	   * starting at list head scan
	   * for locks in the range by
	   * this process
	   */
	  cl = (struct locklist *)&ip->i_locklist; /* note addr is pointer in inode */
	  while(nl = cl->ll_link) {
	       /*
		* if not by this process skip to next lock
		*/
	       if(nl->ll_proc != u.u_procp) {
		    cl = nl;
		    continue;
	       }
	       /*
		* check for locks in proper range
		*/
	       if( UB <= nl->ll_start )
		    break;
	       if( nl->ll_end <= LB ) {
		    cl = nl;
		    continue;
	       }
	       /*
		* for locks fully contained within
		* requested range, just delete the item
		*/
	       if( LB <= nl->ll_start && nl->ll_end <= UB) {
		    cl->ll_link = nl->ll_link;
		    lockfree(nl);
		    continue;
	       }
	       /*
		* if some one is sleeping on this lock
		* do a wakeup, we may free the region
		* being slept on
		*/
	       if(nl->ll_flags & IWANT) {
		    nl->ll_flags &= ~IWANT;
		    wakeup((caddr_t) nl);			/* M001 */
	       }
	       /*
		* middle section is being removed
		* add new lock for last section
		* modify existing lock for first section.
		* if no locks, return in error
		*/
	       if( nl->ll_start < LB && UB < nl->ll_end) {
		    if( lockadd(nl,nl->ll_flags&LROLOCK,UB,nl->ll_end) ) return; /*M000*/
		    nl->ll_end = LB;
		    break;
	       }
	       /*
		* first section is being deleted
		* just move starting point up
		*/
	       if( LB <= nl->ll_start && UB < nl->ll_end) {
		    nl->ll_start = UB;
		    break;
	       }
	       /*
		* must be deleting last part of this section
		* move ending point down
		* continue looking for locks covered by upper
		* limit of unlock range
		*/
	       nl->ll_end = LB;
	       cl = nl;
	  }
	  /*
	   * end of scan for unlock request
	   */
	  return;
     }
     /*
      * request must be a lock of some kind
      * check to see if the region is lockable by this
      * process
      */
     if(locked(nblock,ip,LB,UB))        /* M000 if the request is nonblocking, */
	return;                         /* M000 and is locked, return */
     cl = (struct locklist *)&ip->i_locklist; /* note addr is pointer in inode */
     /*
      * simple case, no existing locks, simply add new lock
      */
     if( (nl=cl->ll_link) == NULL ) {
	  lockadd(cl,rolock,LB,UB);     /*M000*/
	  return;
     }
     /*
      * simple case, lock is before existing locks,
      * simply insert at head of list
      */
     if( UB < nl->ll_start ) {
	  lockadd(cl,rolock,LB,UB);     /*M000*/
	  return;
     }
     /*
      * ending range of lock is same as start of lock by
      * another process, simply insert at head of list
      */

     if( UB <= nl->ll_start && u.u_procp != nl->ll_proc ) {
	  lockadd(cl,rolock,LB,UB);     /*M000*/
	  return;
     }
     /*
      * request range overlaps with begining of first request
      * if locks are of the same type, simply
      * modify starting point in lock to include requested region
      */
     if( UB >= nl->ll_start && LB < nl->ll_start ) {
	  /*
	   * locks overlap and are of different types
	   */
	  if((nl->ll_flags&LROLOCK) != (rolock&LROLOCK)) {          /* M000 */
		/*
		 * end of new lock is within range of old lock
		 * modify start of old lock to exclude newly locked region
		 * and add new lock
		 */
		if(UB < nl->ll_end) {                   /* M000 */
			nl->ll_start = UB;              /* M000 */
			lockadd(cl,rolock,LB,UB);       /* M000 */
			return;                         /* M000 */
		}                                       /* M000 */
		/*
		 * new lock fully includes old one
		 * change lock's flag and LB values for new lock
		 */
		if(UB == nl->ll_end) {                  /* M000 */
			nl->ll_start = LB;              /* M000 */
			nl->ll_flags = rolock;          /* M000 */
			return;                         /* M000 */
		}                                       /* M000 */
	  }                                             /* M000 */

	  nl->ll_start = LB;
     }
     /*
      * scan thru remaining locklist
      */
     cl = nl;
     while(1) {
	  /*
	   * actions for requests at end of list
	   */
	  if( ( nl = cl->ll_link ) == NULL ) {
	       /*
		* request overlaps tail of last entry
		* check for compatible lock types and either extend end point
		* or create new lock
		*/
	       if( LB <= cl->ll_end && u.u_procp == cl->ll_proc ) {
		    if((cl->ll_flags&LROLOCK) != (rolock&LROLOCK)) {        /* M000 */
			if(LB > cl->ll_start) {                             /* M000 */
				if(UB < cl->ll_end)                         /* M000 */
				     lockadd(cl,(int)cl->ll_flags,UB,cl->ll_end);
				cl->ll_end = LB;                /* M000 */
				lockadd(cl,rolock,LB,UB);       /* M000 */
				return;                         /* M000 */
			 }                                      /* M000 */
			 if(UB < cl->ll_end) {                  /* M000 */
				oldf = cl->ll_flags;            /* M000 */
				oldend = cl->ll_end;            /* M000 */
				cl->ll_flags = rolock!=0 ? LROLOCK : 0;
				cl->ll_end = UB;                /* M000 */
				lockadd(cl,oldf,UB,oldend);     /* M000 */
				return;                         /* M000 */
			 }                                      /* M000 */
			 cl->ll_end = UB;                       /* M000 */
			 cl->ll_flags = rolock!=0 ? LROLOCK : 0;/* M000 */
			 return;                                /* M000 */
		    }
		    if( UB > cl->ll_end ) cl->ll_end = UB;
		    return;
	       }
	       /*
		* otherwise add new entry
		*/
	       lockadd(cl,rolock,LB,UB);        /*M000*/
	       return;
	  }
	  /*
	   * if more locks in range skip to next
	   * otherwise stop scan
	   */
	  if( nl->ll_start < LB ) {
	       cl = nl;
	  }
	  else {
	       break;
	  }
     }
     /*
      * if upper bound is fully resolved and locking mode is the same
      * were done, otherwise fix end of last entry or add new entry
      */
     if(UB <= cl->ll_end) {                                     /* M000 */
	if((cl->ll_flags&LROLOCK) != (rolock&LROLOCK)) {        /* M000 */
	    if(LB > cl->ll_start) {                             /* M000 */
	       if(UB < cl->ll_end) lockadd(cl,(int)cl->ll_flags,UB,cl->ll_end);
	       cl->ll_end = LB;                                 /* M000 */
	       lockadd(cl,rolock,LB,UB);                        /* M000 */
	       return;
	    }
	    if(LB == cl->ll_start){                             /* M000 */
	       if(UB < cl->ll_end) {
		  lockadd(cl,(int)cl->ll_flags,UB,cl->ll_end);
		  cl->ll_end = UB;                              /* M000 */
		  cl->ll_flags = rolock != 0 ? LROLOCK : 0;     /* M000 */
		  return;
	       }
	       cl->ll_flags = rolock != 0 ? LROLOCK : 0;        /* M000 */
	    }
	} else return;                                          /* M000 */
     }
     if(LB <= cl->ll_end && u.u_procp == cl->ll_proc) {         /* M000 */
	if((cl->ll_flags&LROLOCK) != (rolock&LROLOCK)) {        /* M000 */
		if(LB > cl->ll_start) {                         /* M000 */
			cl->ll_end = LB;                        /* M000 */
			lockadd(cl,rolock,LB,UB);               /* M000 */
			cl = cl->ll_link;                       /* M000 */
		} else {                                        /* M000 */
			cl->ll_end = UB;                        /* M000 */
			cl->ll_flags = rolock;                  /* M000 */
		}                                               /* M000 */
	} else cl->ll_end = UB;                                 /* M000 */
     } else {
	  if( lockadd(cl,rolock,LB,UB) ) return;        /*M000*/
	  cl = cl->ll_link;
     }
     /*
      * end point set above may overlap later entries
      * if so delete or modify them to perform the compaction
      */
     while( (nl=cl->ll_link) != NULL)  {
	  /*
	   * if we found lock by another process we must
	   * be done since we validated the range above
	   */
	  if(u.u_procp != nl->ll_proc) return;
	  /*
	   * if the new endpoint no longer overlaps were done
	   */
	  if(cl->ll_end < nl->ll_start) return;
	  /*
	   * if the new range overlaps the first part of the
	   * next lock, take its end point
	   * and delete the next lock
	   * we should be done
	   */
	  if(cl->ll_end <= nl->ll_end) {
		if((cl->ll_flags&LROLOCK) != (nl->ll_flags&LROLOCK)) {    /* M000 */
			if(nl->ll_end > UB) {                   /* M000 */
				nl->ll_start = UB;              /* M000 */
				return;                         /* M000 */
			}                                       /* M000 */
		} else {                                        /* M000 */
			cl->ll_end = nl->ll_end;
			cl->ll_link = nl->ll_link;
			lockfree(nl);
			return;
		}
	  }
	  /*
	   * the next lock is fully included in the new range
	   * so it may be deleted
	   */
	  cl->ll_link = nl->ll_link;
	  lockfree(nl);
     }
}
/*
 * locked -- routine to scan locks and check for a locked condition
 *    returns 0 if area not locked, non-zero if is.  The value of 'flag'
 *    determines the action when the area is locked.  If 'flag' is 0
 *    the caller is always suspended.  If 'flag' is 1 the caller is only
 *    suspended on a r/w lock.  If 'flag' is 2 the caller is not suspended.
 *
 *        Locked is called by locking and the hooks in sys2.c. It is
 *   passed a pointer to the incore inode structure and a flag which
 *   signifies either error return or sleeping when a locked area is
 *   found.
 *        Locked scans the lock list until the entire requested region
 *   is available without blocking. If an error return is requested, locked
 *   sets the error code and returns if the region is not available.
 *        If sleeping is necessary, the deadlock module is called to see
 *   if sleeping will cause a deadlock -- if so, locked returns an
 *   error.  If sleeping is ok, locked then sleeps on the list item
 *   until it is released.  When the list item is released, locked
 *   rescans the list looking for other locked regions.
 */

locked(flag,ip,LB,UB)
int flag;
register struct inode *ip;
off_t     LB, UB;
{
     register struct locklist *nl = ip->i_locklist;

     /*
      * scan list while locks are in requested range
      */
     while(nl != NULL && nl->ll_start < UB) {
	  /*
	   * skip locks for this process
	   * and those out of range
	   */
	  if( nl->ll_proc == u.u_procp || nl->ll_end <= LB) {
	       nl = nl->ll_link;
	       if(nl == NULL) return(NULL);
	       continue;
	  }
	  /*
	   * must have found lock by another process
	   * if request is to test only, then exit with
	   * error code
	   */
	  if(flag==1)                                           /* M000 */
	       if((nl->ll_flags & LROLOCK) != 0)                /* M000 */
		       return(0);                     /* M000 *//* M001 */
	  if(flag==2) {                                         /* M000 */
	       u.u_error = EACCES;                              /* M000 */
	       return(1);                                       /* M000 */
	  }

	  /*
	   * will need to sleep on lock, check for deadlock first
	   * abort on error
	   */
	  if(deadlock(nl) != NULL) return(1);
	  /*
	   * post want flag to get awoken
	   * then sleep till lock is released
	   */
	  nl->ll_flags |= IWANT;
	  sleep( (caddr_t)nl, PSLEP);
	  /*
	   * set scan back to begining to catch
	   * any new areas locked
	   * or a partial delete
	   */
	  nl = ip->i_locklist;
	  /*
	   * abort if any errors
	   */
	  if(u.u_error) return(1);
     }
     return(0);				/* M001 */
}

/*
 * deadlock -- routine to follow chain of locks and proc table entries
 *        to find deadlocks on file locks.
 *
 *        Deadlock is called from locked to trace back thru the locklist
 *   and proc table for a deadlock. If a deadlock is detected, then it
 *   error returns; otherwise, it returns NULL.
 *        Deadlock is called with a pointer to the list item on which
 *   the sleep will occur. The id entry in that list item is used to see
 *   if that process is currently sleeping on another list item. If the
 *   "other" process is not sleeping on a list item, no deadlock is
 *   pending, and a normal return is made. If the "other" process is
 *   sleeping on a list item, and that item was created by this process,
 *   a deadlock is pending, and error return is made. Otherwise the id
 *   in this new list item is used to check other processes in the chain
 *   until one is found which is either not sleeping on a list item, or
 *   sleeping on a lock by this process.
 */

deadlock(lp)
register struct locklist *lp;
{
     register struct locklist *nl;

     /*
      * scan while the process owning the lock is sleeping
      */
     while(lp->ll_proc->p_stat == SSLEEP) {
	  /*
	   * if the object the process is sleeping on is
	   * NOT in the locktable every thing is ok
	   * fall out of loop and return NULL
	   */
	  nl = (struct locklist *)lp->ll_proc->p_wchan;
	  if( nl < &locklist[0] || nl >= &locklist[v.v_lock] )
	       break;
	  /*
	   * the object was a locklist entry
	   * if the owner of that entry is this
	   * process then a deadlock would occur
	   * set error exit and return
	   */
	  if(nl->ll_proc == u.u_procp) {
	       u.u_error = EDEADLOCK;
	       return(1);
	  }
	  /*
	   * the object was a locklist entry
	   * owned by some other process
	   * continue the scan with that process
	   */
	  lp = nl;
     }
     return(NULL);
}

/*
 * unlckf -- called by close to release all locks for this process
 *
 *        Unlckf is called by closef to release any locks to the
 *   argument inode which are held by the process.  The routine scans
 *   that inode's list, removing all items by that process.
 */

unlckf(ip)
register struct inode *ip;
{
     register struct locklist *nl;
     register struct locklist *cl;

     if(cl = (struct locklist *)&ip->i_locklist)
	  while( (nl = cl->ll_link) != NULL) {
	       if(nl->ll_proc == u.u_procp) {
		    cl->ll_link = nl->ll_link;
		    lockfree(nl);
	       }
	       else cl = nl;
	  }
}

/*
 * lockalloc -- allocates free list, returns free lock items
 * On the first call lockalloc initializes the free lock list.
 * Lockalloc returns either the next free item on the locklist,
 * or NULL if none are available.
 */

struct locklist *
lockalloc(){
     register struct locklist *fl = &locklist[0];
     register struct locklist *nl;

     /*
      * link the locklist table into the freelist
      */
     if(fl->ll_proc == NULL) {
	  fl->ll_proc = &proc[0];
	  for(nl= &locklist[1]; nl < &locklist[v.v_lock]; nl++) {
	       lockfree(nl);
	  }
     }
     /*
      * if all the locks are used error exit
      */
     if( (nl=fl->ll_link) == NULL) {
	  u.u_error = EDEADLOCK;
	  return(NULL);
     }
     /*
      * return the next lock on the list
      */
     fl->ll_link = nl->ll_link;
     nl->ll_link = NULL;
     return(nl);
}

/*
 * lockfree -- returns a lock item to the free list
 *   Lockfree is called by locking to add a locklist entry back
 *   into the free list. If any process is sleeping on the lock, the
 *   process is awakened.
 */

lockfree(lp)
register struct locklist *lp;
{
     register struct locklist *fl = &locklist[0];

     /*
      * if some process is sleeping on this lock
      * wake them up
      */
     if(lp->ll_flags & IWANT) {
	  lp->ll_flags &= ~IWANT;
	  wakeup((caddr_t) lp);				/* M001 */
     }
     /*
      * add the lock into the free list
      */
     lp->ll_link = fl->ll_link;
     fl->ll_link = lp;
}

/*
 * lockadd -- routine to add item to list
 * Lockadd is called from locking to insert a lockitem into the
 * lock list for an inode.  The arguments and data from the current
 * process are used to fill in the entries for the lock.
 */
lockadd(cl,flag,LB,UB)                  /*M000*/
register struct locklist *cl;
int flag;
off_t LB,UB;
{
      register struct locklist *nl;

     /*
      * get a lock, return if none available
      */
     nl = lockalloc();
     if(nl == NULL) {
	  return(1);
     }
     /*
      * link the new entry into list at current spot
      * fill in the data from the args
      */
     nl->ll_link = cl->ll_link;
     cl->ll_link = nl;
     nl->ll_proc = u.u_procp;
     nl->ll_flags = (flag!=0 ? LROLOCK : 0);   /* M000 */
     nl->ll_start = LB;
     nl->ll_end = UB;
     return(0);
}
