/*
 *	@(#) sys/sem.c 1.2 84/01/02 Copyright Microsoft Corporation, 1983 [Microsoft SID]
 *	@(#) sys/sem.c 2.1 86/09/26 Copyright (c) The Santa Cruz Operation, Inc., 1985, 1986; All rights reserved.
 */

/*
 * THIS FILE CONTAINS CODE WHICH IS DESIGNED TO BE
 * PORTABLE BETWEEN DIFFERENT MACHINE ARCHITECHTURES
 * AND CONFIGURATIONS. IT SHOULD NOT REQUIRE ANY
 * MODIFICATIONS WHEN ADAPTING XENIX TO NEW HARDWARE.
 */


/*
 *  Semaphores are inodes of the special file type IFNAM.
 *  Semaphore inodes contain the current count (i_scount) of the
 *  semaphore and pointers to the head and tail of the list of waiters for
 *  the semaphore (i_headw and i_tailw).
 *  When a process must wait for the resource, it puts its file
 *  structure on the waiting list. By convention, the first element on the
 *  waiting list is the one that currently "owns" the semaphore; i.e., the
 *  the one using the resource governed by the semaphore.
 */

/*			MODIFICATION HISTORY
 * $Log:	sem.c,v $
 * Revision 6.0  87/10/12  17:09:47  iv
 * 	- Checkin to lock down 3.2
 * 
 * Revision 5.1  86/10/01  18:17:10  iv
 * 	- S005 by blf@sco:
 * 	  the owner of a semaphore shouldn't wait for it (should get an error)
 * 
 *	
 *	M001	WFS/RWS	14-Oct-82
 *		-Changed test for 'no file' to test against NFILE instead
 *		 of NOFILE.  This is actually a good argument for typed
 *		 constants...
 *	M002	JJD	05 Dec 82
 *		- removed custf field from all argument structures; the
 *		  current cxenix code takes care of this for you.  Uncommented.
 *		- adapted for configurable kernel
 *	M003	JJD	12 Feb 83
 *		- added casts for lint
 *	M004	EA	24 May 83
 *		- removed code in closesem that set an error flag in the
 *		  semaphore inode upon the last close.
 *	26 Sept 1986	blf	S005
 *		- The owner of a semaphore shouldn't wait for it.
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/var.h"			/*M002*/

#define SERROR  01              /* process controlling a sem terminated */
#define isem i_fdep.i_namef.i_ndata.i_sem   /* shorthand notation */


#ifdef DEBUG
int debugsem = 0;
struct file *h;
#endif

/*
 * creatsem system call
 *
 * Creates an instance of a semaphore named sem_name.
 * Semaphores are files of 0 length - the file name space
 * is used to provide unique identifiers for semaphores (the inode number).
 * creatsem fails if a file named sem_name already exists and has been
 * opened by at least one process.
 */

creatsem()
{
	register struct inode *ip;
	register struct a {
	    char    *sem_name;  /* path name specifying the semaphore */
	    int     mode;       /* protection of file */
	} *uap;
	struct file *fp;
	extern  uchar();

	uap = (struct a *)u.u_ap;
	u.u_dirp = uap->sem_name;
	ip = namei(uchar, 1);            /* convert pathname into inode ptr */

	if(ip == NULL) {
		if(u.u_error)
		    return;
		ip = maknode(IFNAM | (uap->mode&(~IFMT)));
		if (ip == NULL)
		    return;
		ip->i_fdep.i_namef.i_type = IFSEM;
	} else {                                /* sem_name already exists */
						/* must not be currently open */
		if (ip->i_count != 1) {
			u.u_error = EEXIST;
			iput(ip);
			return;
		}                               /* must be sem type inode */
		if (((ip->i_mode & IFMT) != IFNAM)
		  || (ip->i_fdep.i_namef.i_type != IFSEM)){
			u.u_error = ENOTNAM;
			iput(ip);
			return;
		}
		else
			access(ip, IREAD);
		}
	prele(ip);
	if ((fp = falloc()) == NULL){
		iput(ip);
		return;
	}
	fp->f_flag = 0;
	fp->f_inode = ip;
	ip->isem.i_scount = 1;
	ip->isem.i_headw = (ip->isem.i_tailw = NULL);
	ip->isem.i_eflag = 0;
	ip->i_flag |= IACC;
}

/*
 * opensem system call
 *
 * Opens a semaphore named sem_name and returns that semaphore's unique
 * identification number.
 * opensem fails if the semaphore doesn't exist, or if a process that
 * controlled the sem terminated without relinquishing control leaving
 * the resource governed by the sem in an inconsistent state (ENAVAIL).
 */

opensem()
{
	register struct inode *ip;
	register struct a {
	    char    *sem_name;      /* path name specifying the semaphore */
	} *uap;
	extern uchar();
	struct file *fp;

	uap = (struct a *)u.u_ap;

	u.u_dirp = uap->sem_name;
	ip = namei(uchar, 0);       /* convert pathname into inode pointer */

	if(ip == NULL)
		return;
				    /* must be sem type inode */
	if(((ip->i_mode&IFMT) != IFNAM) || (ip->i_fdep.i_namef.i_type != IFSEM)){
	    u.u_error = ENOTNAM;
	    iput(ip);
	    return;
	}
	if(ip->isem.i_eflag == SERROR) {
	    u.u_error = ENAVAIL;
	    iput(ip);
	    return;
	    }
	access(ip, IREAD);
	prele(ip);
	if ((fp = falloc()) == NULL){
		iput(ip);
		return;
	}
	fp->f_flag = 0;
	fp->f_inode = ip;
}

/*  sigsem system call */

sigsem(){
	register struct inode *ip;
	register struct file  *fp;
	struct a {
	    int fdes;     /* sem # to signal (special file desc.) */
	} *uap;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);

	if (fp == NULL){
	    u.u_error = EBADF;
	    return;
	}
	ip = fp->f_inode;
					/* must be sem type inode */
	if(((ip->i_mode&IFMT) != IFNAM) || (ip->i_fdep.i_namef.i_type != IFSEM)){
	    u.u_error = ENOTNAM;
	    return;
	}
	plock(ip);
					/* ensure sem owner is signalling */
	if (ip->isem.i_headw != fp) {
	    u.u_error = ENAVAIL;
	    goto out;
	}                               /* remove self from head of list as owner */

#ifdef DEBUG
	if (debugsem) {
		printf("ENTER SIGSEM: ip = %x, scount = %d, eflag = %d, tail = %x\n",
		    ip, ip->isem.i_scount, ip->isem.i_eflag, ip->isem.i_tailw);
		printf("waiters: ");
		for (h = ip->isem.i_headw; h != NULL; h = h->f_un.f_slnk) {
			printf("  %x", h);
		}
		printf("\n");
	}
#endif

	ip->isem.i_headw = fp->f_un.f_slnk;
					/* signal 1st waiting process, if there is one */
	if((ip->isem.i_scount)++ < 0)
	    wakeup((caddr_t)ip->isem.i_headw);    /*M003*/
	ip->i_flag |= IACC;

#ifdef DEBUG
	if (debugsem) {
		printf("EXIT SIGSEM: ip = %x, scount = %d, eflag = %d, tail = %x\n",
		    ip, ip->isem.i_scount, ip->isem.i_eflag, ip->isem.i_tailw);
		printf("waiters: ");
		for (h = ip->isem.i_headw; h != NULL; h = h->f_un.f_slnk) {
			printf("  %x", h);
		}
		printf("\n");
	}
#endif

out:    prele(ip);
	return;
}

/*  waitsem system call */

waitsem(){
	cwaitsem(0);
}

/*  nbwaitsem system call */

nbwaitsem(){
	cwaitsem(1);
}

/*  common code for waitsem and nbwaitsem */

cwaitsem(nowait)
    int nowait;
{
    register struct inode *ip;
    register struct file *fp, *wp;				/* S005 */
    struct a {
	int fdes;     /* sem # to wait on (special file desc.) */
    } *uap;

	uap = (struct a *)u.u_ap;
	fp = getf(uap->fdes);

	if (fp == NULL){
	    u.u_error = EBADF;
	    return;
	}
	ip = fp->f_inode;
					 /* must be sem type inode */
	if(((ip->i_mode&IFMT) != IFNAM) || (ip->i_fdep.i_namef.i_type != IFSEM)){
	    u.u_error = ENOTNAM;
	    return;
	}
	if (ip->isem.i_eflag == SERROR) {
	    u.u_error = ENAVAIL;
	    return;
	}              
	plock(ip);
	/* S005
	 *  Check that this process neither owns the semaphore nor waits
	 *  for it (ie, prevent 2 waitsem calls without an intervening
	 *  sigsem by the same process using the same file descriptor).
	 */
	for (wp = ip->isem.i_headw; wp != NULL; wp = wp->f_un.f_slnk) {
		if (wp == fp) {
			prele(ip);
			u.u_error = EINVAL;
			return;
		}
	}
	if (nowait && ip->isem.i_scount <= 0){	/* nbwait returns error */
	    prele(ip);                                /* instead of pending */
	    u.u_error = ENAVAIL;
	    return;
	}
	ip->i_flag |= IACC;

#ifdef DEBUG
	if (debugsem) {
		printf("ENTER WAITSEM: ip = %x, scount = %d, eflag = %d, tail = %x\n",
		    ip, ip->isem.i_scount, ip->isem.i_eflag, ip->isem.i_tailw);
		printf("waiters: ");
		for (h = ip->isem.i_headw; h != NULL; h = h->f_un.f_slnk) {
			printf("  %x", h);
		}
		printf("\n");
	}
#endif
	fp->f_un.f_slnk = NULL;
	if(--(ip->isem.i_scount) < 0) {   /* sem busy, must wait */
					   /* insert at tail of waiter list */
		ip->isem.i_tailw->f_un.f_slnk = fp;
		ip->isem.i_tailw = fp;
		prele(ip);
		while(fp != ip->isem.i_headw) {
			sleep((caddr_t) fp,PSLEP);	/*M003*/
				/* check if process controlling the sem has */
				/* ceased to use it before giving up control*/
			if(ip->isem.i_eflag == SERROR) {
				u.u_error = ENAVAIL;
				return;
			}
		}
	} else {			   /* insert on list as sem "owner" */
		ip->isem.i_headw = (ip->isem.i_tailw = fp);
		prele(ip);
	}
#ifdef DEBUG
	if (debugsem) {
		printf("EXIT WAITSEM: ip = %x, scount = %d, eflag = %d, tail = %x\n",
		    ip, ip->isem.i_scount, ip->isem.i_eflag, ip->isem.i_tailw);
		printf("waiters: ");
		for (h = ip->isem.i_headw; h != NULL; h = h->f_un.f_slnk) {
			printf("  %x", h);
		}
		printf("\n");
	}
#endif
}

/*
 * closesem is called from closef to cleanup in case a terminating process
 * is the current "owner" of a semaphore protected resource or is waiting
 * on a semaphore.
 */

closesem(fp,ip)
    struct file *fp;
    register struct inode *ip;
{
    register struct file *lfp, *tfp;

    lfp = NULL;
    for(tfp=ip->isem.i_headw; tfp != NULL; lfp=tfp, tfp=tfp->f_un.f_slnk){
	if(tfp == fp) {
	   if (lfp == NULL) {           /* process is owner of semaphore */
		ip->isem.i_eflag = SERROR;
		ip->isem.i_scount++;
					/* cause all processes waiting to error return */
		for(tfp = tfp->f_un.f_slnk; tfp != NULL; tfp = tfp->f_un.f_slnk) {
			ip->isem.i_scount++;
			wakeup((caddr_t) tfp);		/*M003*/
		}
		ip->i_flag |= ICHG;
		ip->isem.i_headw = NULL;
		return;
	    }
					/* remove self from waiting list */
	    ip->isem.i_scount++;
	    lfp->f_un.f_slnk = tfp->f_un.f_slnk;
	    if (tfp->f_un.f_slnk == NULL)
		ip->isem.i_tailw = lfp;
	    ip->i_flag |= ICHG;
	}
    }
}		/* M004 */

/* passem handles inheritance of semaphores over forks.
 * Each child is given its own copy of a file structure referencing
 * the semaphore.
 */

passem()
{
	register struct file *fp;
	register int i;

	for (i = 0; i < NOFILE; i++) {
	    if (u.u_ofile[i]){

			     /* if an open semaphore was inherited */
		if (((u.u_ofile[i]->f_inode->i_mode & IFMT) == IFNAM) &&
		     (u.u_ofile[i]->f_inode->i_fdep.i_namef.i_type == IFSEM)){

				  /* replace the inherited file structure */
				  /* with a new one not shared with parent */
		    for (fp = &file[0]; fp < v.ve_file; fp++) {	/*M002*/
			if (fp->f_count == 0) {    /* grab new file structure */
			    fp->f_count++;
			    fp->f_offset = 0;
			    fp->f_inode = u.u_ofile[i]->f_inode;
			    fp->f_inode->i_count++;
			    u.u_ofile[i]->f_count--;
				    /* release old file structure */
			    u.u_ofile[i] = fp;    
				    /* ensure shared file descriptor */
				    /* references different file structure */
			    break; 
			}
		    }
		    if (fp >= v.ve_file) {	/* M001	*/	/*M002*/
		       printf("no file\n");
		       u.u_error = ENFILE;
		       return(-1);
		    }
		}
	    }
	}
	return(0);
}


